import asyncio
import re
import shutil
from dataclasses import dataclass
from pathlib import Path
from textwrap import dedent

import anyio

from clypi import Command, Positional, Spinner, arg, boxed, print
from clypi.colors import style

MDTEST_DIR = Path.cwd() / ".mdtest"
PREAMBLE = """\
import clypi
import clypi.parsers as cp
from pathlib import Path
from typing import reveal_type
from clypi import *
from enum import Enum
"""


class TestFailed(Exception):
    pass


@dataclass
class Test:
    name: str
    orig: str
    code: str
    args: str
    stdin: str


async def parse_file(file: Path) -> list[Test]:
    tests: list[Test] = []
    base_name = (
        file.relative_to(Path.cwd())
        .as_posix()
        .replace("/", "-")
        .replace(".md", "")
        .lower()
    )

    async with await anyio.open_file(file, "r") as f:
        in_test, current_test, args, stdin = False, [], "", ""
        async for line in f:
            # End of a code block
            if "```" in line and current_test:
                code = "\n".join(current_test[1:])
                tests.append(
                    Test(
                        name=f"{base_name}-{len(tests)}",
                        orig=dedent(code),
                        code=PREAMBLE + dedent(code),
                        args=args,
                        stdin=stdin + "\n",
                    )
                )
                in_test, current_test, args, stdin = False, [], "", ""

            # We're in a test, accumulate all lines
            elif in_test:
                current_test.append(line.removeprefix("> ").rstrip())

            # Mdtest arg definition
            elif g := re.search("<!--- mdtest-args (.*) -->", line):
                args = g.group(1)
                in_test = True

            # Mdtest stdin definition
            elif g := re.search("<!--- mdtest-stdin (.*) -->", line):
                stdin = g.group(1)
                in_test = True

            # Mdtest generic definition
            elif g := re.search("<!--- mdtest -->", line):
                in_test = True

    return tests


async def run_test(test: Test) -> tuple[str, str]:
    # Save test to temp file
    test_file = MDTEST_DIR / f"{test.name}.py"
    test_file.write_text(test.code)

    # Run the test
    file_rel = test_file.relative_to(Path.cwd())
    command = f"uv run --all-extras {file_rel}"
    if test.args:
        command += f" {test.args}"

    # Await the subprocess to run it
    proc = await asyncio.create_subprocess_shell(
        command,
        stdout=asyncio.subprocess.PIPE,
        stdin=asyncio.subprocess.PIPE,
        stderr=asyncio.subprocess.PIPE,
    )
    stdout, stderr = await proc.communicate(test.stdin.encode())

    # If no errors, return
    if proc.returncode == 0:
        return test.name, ""

    # If there was an error, pretty print it
    error = []
    error.append(style(f"\n\nError running test {test.name!r}\n", fg="red", bold=True))
    error.append(boxed(test.orig, title="Code"))

    if stdout.decode():
        error.append("")
        error.append(boxed(stdout.decode().strip(), title="Stdout"))

    if stderr.decode():
        error.append("")
        error.append(boxed(stderr.decode().strip(), title="Stderr"))

    return test.name, "\n".join(error)


async def run_mdtests(tests: list[Test]):
    errors = []
    async with Spinner("Running Markdown Tests") as s:
        coros = [run_test(test) for test in tests]
        for task in asyncio.as_completed(coros):
            idx, err = await task
            if not err:
                s.log(style("✔", fg="green") + f" Finished test {idx}")
            else:
                errors.append(err)
                s.log(style("×", fg="red") + f" Finished test {idx}")

        if errors:
            await s.fail()

    for err in errors:
        print(err)


class Mdtest(Command):
    """
    Run python code embedded in markdown files to ensure it's
    runnable.
    """

    files: Positional[list[Path]] = arg(
        help="The list of markdown files to test",
        default_factory=lambda: list(Path.cwd().glob("**/*.md")),
    )

    async def run(self) -> None:
        # Setup test dir
        MDTEST_DIR.mkdir(exist_ok=True)

        # Assert each file exists
        for file in self.files:
            assert file.exists(), f"File {file} does not exist!"

        # Collect tests
        async with Spinner("Collecting Markdown Tests"):
            per_file = await asyncio.gather(*(parse_file(file) for file in self.files))
            all_tests = [test for file in per_file for test in file]

        # Run each file
        await run_mdtests(all_tests)

        # Cleanup
        shutil.rmtree(MDTEST_DIR)


if __name__ == "__main__":
    mdtest = Mdtest.parse()
    mdtest.start()



import inspect
import typing as t
from enum import Enum
from types import UnionType

P = t.ParamSpec("P")
R = t.TypeVar("R")


def ignore_annotated(fun: t.Callable[P, R]) -> t.Callable[P, R]:
    def inner(*args: P.args, **kwargs: P.kwargs) -> R:
        if t.get_origin(args[0]) == t.Annotated:
            args_ls = list(args)
            args_ls[0] = t.get_args(args[0])[0]
            args = tuple(args_ls)  # type: ignore
        return fun(*args, **kwargs)

    return inner


@ignore_annotated
def is_list(_type: t.Any) -> t.TypeGuard[list[t.Any]]:
    return t.get_origin(_type) in (list, t.Sequence)


@ignore_annotated
def list_inner(_type: t.Any) -> t.Any:
    return t.get_args(_type)[0]


@ignore_annotated
def is_tuple(_type: t.Any) -> t.TypeGuard[tuple[t.Any]]:
    return t.get_origin(_type) is tuple


@ignore_annotated
def tuple_inner(_type: t.Any) -> tuple[list[t.Any], int | None]:
    """
    Returns the list of types for the tuple and how many items
    it accepts
    """
    # TODO: can be made more efficient
    inner_types = list(t.get_args(_type))
    if inner_types[-1] is Ellipsis:
        return [inner_types[0]], None
    return inner_types, len(inner_types)


@ignore_annotated
def is_union(_type: t.Any) -> t.TypeGuard[UnionType]:
    return isinstance(_type, UnionType)


@ignore_annotated
def union_inner(_type: t.Any) -> list[t.Any]:
    return list(t.get_args(_type))


@ignore_annotated
def is_literal(_type: t.Any) -> bool:
    return t.get_origin(_type) == t.Literal


@ignore_annotated
def literal_inner(_type: t.Any) -> list[t.Any]:
    return list(t.get_args(_type))


@ignore_annotated
def tuple_size(_type: t.Any) -> float:
    args = _type.__args__
    if args[-1] is Ellipsis:
        return float("inf")
    return len(args)


@ignore_annotated
def is_none(_type: t.Any) -> t.TypeGuard[type[None]]:
    return _type is None


@ignore_annotated
def is_enum(_type: t.Any) -> t.TypeGuard[type[Enum]]:
    return inspect.isclass(_type) and issubclass(_type, Enum)


@ignore_annotated
def remove_optionality(_type: t.Any) -> t.Any:
    if not isinstance(_type, UnionType):
        return _type

    new_args: list[t.Any] = []
    for arg in _type.__args__:
        if arg is not type[None]:
            new_args.append(arg)

    if len(new_args) == 1:
        return new_args[0]

    return t.Union[*new_args]


@ignore_annotated
def type_to_str(_type: t.Any) -> str:
    _map = {
        "bool": "boolean",
        "int": "integer",
        "float": "float",
        "str": "string",
        "Path": "Path",
    }
    if inspect.isclass(_type) and _type.__name__ in _map:
        return _map[_type.__name__]

    if t.get_origin(_type) is t.Literal:
        return "{" + "|".join(type_to_str(tp) for tp in _type.__args__) + "}"

    if inspect.isclass(_type) and issubclass(_type, Enum):
        return "{" + "|".join(tp.name for tp in _type) + "}"

    if isinstance(_type, UnionType):
        return "[" + "|".join(type_to_str(tp) for tp in _type.__args__) + "]"

    if is_tuple(_type):
        args = t.get_args(_type)
        if args[-1] is Ellipsis:
            return "(" + type_to_str(args[0]) + ", ...)"
        else:
            inner = ", ".join(type_to_str(tp) for tp in args)
            maybe_comma = "," if len(args) == 1 else ""  # E.g.: ("foo",)
            return "(" + inner + maybe_comma + ")"

    if is_list(_type):
        return "list[" + ", ".join(type_to_str(tp) for tp in t.get_args(_type)) + "]"

    return str(_type)



from __future__ import annotations

import enum
import re
import typing as t
from abc import ABC, abstractmethod
from datetime import datetime, timedelta
from pathlib import Path as _Path

from typing_extensions import override

from clypi._cli import type_util as tu

T = t.TypeVar("T", covariant=True)
X = t.TypeVar("X")
Y = t.TypeVar("Y")


class UnparseableException(Exception):
    pass


class Parser(t.Protocol[T]):
    def __call__(self, raw: str | list[str], /) -> T: ...


class CannotParseAs(Exception):
    def __init__(self, value: t.Any, parser: Parser) -> None:
        message = f"Cannot parse {value!r} as {parser}"
        super().__init__(message)


class ClypiParser(ABC, t.Generic[X]):
    @abstractmethod
    def __call__(self, raw: str | list[str], /) -> X: ...

    def _repr_args(self, args: list[tuple[str, t.Any]]) -> str | None:
        return None

    def __repr__(self):
        name = self.__class__.__name__.lower()
        args = [(k, v) for k, v in vars(self).items() if not k.startswith("_")]
        args_str = self._repr_args(args)
        if args_str is None:
            return name
        return f"{name}({args_str})"

    def __or__(self, other: ClypiParser[Y]) -> Union[X, Y]:
        return Union(self, other)

    def __eq__(self, other: t.Any):
        if not isinstance(other, ClypiParser):
            return False
        if self.__class__ != other.__class__:
            return False
        # TODO: fix this comparison. Only for tests so it's ok for now
        return str(self) == str(other)


class Int(ClypiParser[int]):
    def __call__(self, raw: str | list[str], /) -> int:
        if isinstance(raw, list):
            raise CannotParseAs(raw, self)
        if int(raw) != float(raw):
            raise ValueError(f"The value {raw!r} is not a valid integer.")
        return int(raw)


class Float(ClypiParser[float]):
    def __call__(self, raw: str | list[str], /) -> float:
        if isinstance(raw, list):
            raise CannotParseAs(raw, self)
        return float(raw)


class Bool(ClypiParser[bool]):
    TRUE_BOOL_STR_LITERALS: set[str] = {"true", "yes", "y"}
    FALSE_BOOL_STR_LITERALS: set[str] = {"false", "no", "n"}

    def __call__(self, raw: str | list[str], /) -> bool:
        if isinstance(raw, list):
            raise CannotParseAs(raw, self)

        raw_lower = raw.lower()
        both = self.TRUE_BOOL_STR_LITERALS | self.FALSE_BOOL_STR_LITERALS
        if raw_lower not in both:
            raise ValueError(
                f"The string {raw!r} is not valid boolean! The only allowed values are: {both}."
            )
        return raw_lower in self.TRUE_BOOL_STR_LITERALS


class Str(ClypiParser[str]):
    def __call__(self, raw: str | list[str], /) -> str:
        if isinstance(raw, list):
            raise CannotParseAs(raw, self)
        return raw


class DateTime(ClypiParser[datetime]):
    def __call__(self, raw: str | list[str], /) -> datetime:
        from dateutil.parser import parse

        if isinstance(raw, list):
            raise CannotParseAs(raw, self)
        return parse(raw)


class TimeDelta(ClypiParser[timedelta]):
    TIMEDELTA_UNITS = {
        ("weeks", "week", "w"): "weeks",
        ("days", "day", "d"): "days",
        ("hours", "hour", "h"): "hours",
        ("minutes", "minute", "m"): "minutes",
        ("seconds", "second", "s"): "seconds",
        ("milliseconds", "millisecond", "ms"): "milliseconds",
        ("microseconds", "microsecond", "us"): "microseconds",
    }
    TIMEDELTA_REGEX = re.compile(r"^(\d+)\s*(\w+)$")

    def __call__(self, raw: str | list[str], /) -> timedelta:
        if isinstance(raw, timedelta):
            return raw

        if not isinstance(raw, str):
            raise ValueError(
                f"Cannot parse {raw!r} as timedelta. Expected str or timedelta, got {type(raw).__name__}"
            )

        match = self.TIMEDELTA_REGEX.match(raw)
        if match is None:
            raise ValueError(f"Invalid timedelta {raw!r}.")

        value, unit = match.groups()
        for units in self.TIMEDELTA_UNITS:
            if unit in units:
                return timedelta(**{self.TIMEDELTA_UNITS[units]: int(value)})

        raise ValueError(f"Invalid timedelta {raw!r}.")


class Path(ClypiParser[_Path]):
    def __init__(self, *, exists: bool = False) -> None:
        self.exists = exists

    def __call__(self, raw: str | list[str], /) -> _Path:
        if isinstance(raw, list):
            raise CannotParseAs(raw, self)
        p = _Path(raw)

        # Validations on the path
        if self.exists and not p.exists():
            raise ValueError(f"File {p.resolve()} does not exist!")

        return p

    @override
    def _repr_args(self, args: list[tuple[str, t.Any]]) -> str | None:
        return ", ".join(f"{k}={v}" for k, v in args if v) or None


class List(ClypiParser[list[X]]):
    def __init__(self, inner: Parser[X]) -> None:
        self._inner = inner

    def __call__(self, raw: str | list[str], /) -> list[X]:
        if isinstance(raw, str):
            raw = raw.split(",")
        return [self._inner(item) for item in raw]

    @override
    def _repr_args(self, args):
        return str(args[0])


class Tuple(ClypiParser[tuple[t.Any]]):
    def __init__(self, *inner: Parser[t.Any], num: int | None = None) -> None:
        self._inner = list(inner)
        self._num = num

    # TODO: can we return the right type here?
    def __call__(self, raw: str | list[str], /) -> tuple[t.Any, ...]:
        if isinstance(raw, str):
            raw = raw.split(",")

        if self._num and len(raw) != self._num:
            raise ValueError(
                f"Expected tuple of length {self._num} but instead got {len(raw)} items"
            )

        # Get all parsers for each item in the tuple (or reuse if tuple[T, ...])
        if not self._num:
            inner_parsers = [self._inner[0]] * len(raw)
        else:
            inner_parsers = self._inner

        # Parse each item with it's corresponding parser
        return tuple(parser(raw_item) for parser, raw_item in zip(inner_parsers, raw))

    @override
    def _repr_args(self, args: list[tuple[str, t.Any]]) -> str | None:
        return ", ".join(str(it) for it in self._inner)


class Union(ClypiParser[t.Union[X, Y]]):
    def __init__(self, left: Parser[X], right: Parser[Y]) -> None:
        self._left = left
        self._right = right

    def __call__(self, raw: str | list[str], /) -> t.Union[X, Y]:
        try:
            return self._left(raw)
        except Exception:
            return self._right(raw)

    def _parts(self):
        """
        Some recursive magic here to "flatten" unions
        """
        left = self._left._parts() if isinstance(self._left, Union) else str(self._left)
        right = (
            self._right._parts() if isinstance(self._right, Union) else str(self._right)
        )
        return f"{left}|{right}"

    def __repr__(self):
        return "(" + self._parts() + ")"


class Literal(ClypiParser[t.Any]):
    def __init__(self, *values: t.Any) -> None:
        self._values = values

    # TODO: can we return the right type here?
    def __call__(self, raw: str | list[str], /) -> t.Any:
        for v in self._values:
            if v == raw:
                return v
        raise CannotParseAs(raw, self)

    def __repr__(self):
        values = "|".join(str(v) for v in self._values)
        return "{" + values + "}"


class NoneParser(ClypiParser[None]):
    def __call__(self, raw: str | list[str], /) -> t.Any:
        raise UnparseableException()


class Enum(ClypiParser[type[enum.Enum]]):
    def __init__(self, _type: type[enum.Enum]) -> None:
        self._type = _type

    def __call__(self, raw: str | list[str], /) -> t.Any:
        if not isinstance(raw, str):
            raise CannotParseAs(raw, self)

        for enum_val in self._type:
            if raw.lower() == enum_val.name.lower():
                return enum_val

        raise ValueError(f"Value {raw} is not a valid choice between {self}")

    def __repr__(self):
        values = "|".join(v.name.lower() for v in self._type)
        return "{" + values + "}"


@tu.ignore_annotated
def from_type(_type: type) -> Parser[t.Any]:
    if _type is bool:
        return Bool()

    if _type is int:
        return Int()

    if _type is float:
        return Float()

    if _type is str:
        return Str()

    if _type is _Path:
        return Path()

    if _type is datetime:
        return DateTime()

    if _type is timedelta:
        return TimeDelta()

    if tu.is_list(_type):
        inner = from_type(tu.list_inner(_type))
        return List(inner)

    if tu.is_tuple(_type):
        inner_types, num = tu.tuple_inner(_type)
        inner_parsers = [from_type(it) for it in inner_types]
        return Tuple(*inner_parsers, num=num)

    if tu.is_union(_type):
        union_inner = tu.union_inner(_type)
        trav = Union(from_type(union_inner[0]), from_type(union_inner[1]))
        for rest in union_inner[2:]:
            trav = Union(trav, from_type(rest))
        return trav

    if tu.is_literal(_type):
        return Literal(*tu.literal_inner(_type))

    if tu.is_none(_type):
        return NoneParser()

    if tu.is_enum(_type):
        return Enum(_type)

    raise UnparseableException(
        f"Don't know how to parse as {tu.type_to_str(_type)} ({type(_type)})"
    )
